#!/usr/bin/perl

package ACL2::Proyecto;

=head1 NAME

ACL2::Proyecto - Procesa proyectos de uno o más ficheros de ACL2

=cut

#
# Procesador de proyectos multifichero de ACL2
# Copyright (C) 2008 Antonio García Domínguez
# $Id$
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use strict;
use warnings;
use utf8;

use ACL2::Dependencias;
use ACL2::Localizacion;
use ACL2::Demostracion;

use IO::File;
use File::stat;
use File::Temp qw(tempfile);
use File::Spec;
use Cwd;
use Carp;

my $lh = ACL2::Localizacion->get_handle();

###################################################
#
# METODOS
#
###################################################

=head2 new

Uso: Constructor.

Entrada: Hash con los siguientes elementos:

=over 2

=item *

ruta_lisp (obligatorio): Ruta relativa al fichero con la fuente Lisp
principal del proyecto.

=item *

acl2_system_books_dir (opcional): Ruta que contiene a todos los libros
del sistema de ACL2. Estos libros no se procesarE<aacute>n.

=item *

deps (opcional): Dependencias de la demostraciE<oacute>n actual. Si no
se proporcionan, se obtendrE<aacute>n de forma automE<aacute>tica.

=back

=cut

sub new
{
    my ($class, %args) = @_;

    my $self = bless {
        "ruta_lisp" => $args{ruta_lisp},
        "acl2_system_books_dir" => $args{acl2_system_books_dir},
        "deps" => $args{deps},
    }, $class;

    if (!$self->{deps})
    {
        $self->{deps} = new ACL2::Dependencias("ruta_lisp" => $self->{ruta_lisp});
    }

    return $self;
}

=head2 get_ruta_lisp

Devuelve: Ruta al fichero Lisp en que consiste esta demostraciE<oacute>n.

=cut

sub get_ruta_lisp
{
    return $_[0]->{ruta_lisp};
}

=head2 get_deps

Devuelve: Objeto ACL2::Dependencias con las dependencias de esta
demostraciE<oacute>n.

=cut

sub get_deps
{
    return $_[0]->{deps};
}

=head2 actualizar

Uso: Genera un documento XML con el postprocesado de la
demostraciE<oacute>n de ACL2, asegurE<aacute>ndose de mantener todas
sus dependencias actualizadas (como en el caso de los libros de
usuario de que dependa).

=cut

sub actualizar
{
    my ($self, %args) = @_;

    my $rutaLISP = $self->get_ruta_lisp;

    my ($volumen, $directorio, $fichLISP) = File::Spec->splitpath($rutaLISP);
    my $basenameLISP = _remove_lisp_extensions($fichLISP);
    my $fichXML      = "$basenameLISP.xml";
    my $fichOUT      = "$basenameLISP.out";
    my $fichACL2     = "$basenameLISP.acl2";
    my $fichACL2Dir  = "cert.acl2";

    # Change current directory to this file's directory
    my $antiguoCWD = getcwd();
    chdir File::Spec->catpath($volumen, $directorio, "")
        or croak $lh->maketext("No se pudo cambiar al directorio de la fuente [_1]", $rutaLISP)
        if $volumen ne "" or $directorio ne "";

    # Protect the rest of the code with an eval so proper cleanup is always done
    my ($dep_cambiadas, $out_desactualizado, $xml_desactualizado);

    eval
    {
        $dep_cambiadas
            = $self->get_deps->actualizar(%args);
        $out_desactualizado
            = $dep_cambiadas
              || $args{force}
              || (-r $fichACL2
                  && __PACKAGE__->_hay_que_actualizar($fichACL2,
                                                      $fichOUT))
              || (-r $fichACL2Dir
                  && __PACKAGE__->_hay_que_actualizar($fichACL2Dir,
                                                      $fichOUT))
              || __PACKAGE__->_hay_que_actualizar($fichLISP, $fichOUT);
        $xml_desactualizado
            = $out_desactualizado
              || __PACKAGE__->_hay_que_actualizar($fichOUT, $fichXML);

        # We will need a decorated temporary .lisp file if we have to
        # update, even if we have the .out file with the original
        # output
        my $pathTmpLISP = _generate_tmp_lisp_file($fichLISP);

        # Do we have to run the proof again?
        if ($out_desactualizado) {
            print STDERR $lh->maketext("Generando [_1]...", $fichOUT) . "\n";

            # Dump ACL2 output to a file so we can analyze it
            my $acl2_status = system("acl2 < ${pathTmpLISP} > $fichOUT");
            croak $lh->maketext(
                "No se pudo escribir la salida de ACL2 de [_1] en [_2]",
                $pathTmpLISP, $fichOUT)
                unless $acl2_status == 0;
        }

        # Do we have to analyze the proof again?
        if ($xml_desactualizado) {
            print STDERR $lh->maketext("Generando [_1]...", $fichXML) . "\n";

            # Should any errors happen, the .xml file should be removed
            eval {
                my $demo = new ACL2::Demostracion(ruta_lisp     => $fichLISP,
                                                  ruta_tmp_lisp => $pathTmpLISP,
                                                  ruta_dem      => $fichOUT,
                                                  deps => $self->get_deps);

                $demo->escribir_xml(destino_xml   => new IO::File(">$fichXML"),
                                    incrustar_dtd => 1);
            };
            if ($@) {
                unlink($fichXML);
                die $@; # Rethrow the error
            }
        }
    };

    # Cleanup: switch back to the previous directory
    chdir $antiguoCWD
        or croak $lh->maketext("No se pudo cambiar de vuelta al directorio [_1]", $antiguoCWD);

    # Dies with any pending errors
    die $@ if $@;

    # Notificamos si ha habido cambios o no
    if ($xml_desactualizado || $out_desactualizado) {
        return 1;
    }
    else {
        return undef;
    }
}

=head2 obtener_xml

Devuelve: CE<oacute>digo XML obtenido de la demostraciE<oacute>n.

=cut

sub obtener_xml
{
    my ($self) = @_;

    my $rutaLISP = $self->get_ruta_lisp;
    my $rutaXML  = _remove_lisp_extensions($rutaLISP) . ".xml";
    return _get_file_contents($rutaXML);
}

=head2 _hay_que_actualizar($rutaFuente, $rutaResultado)

Uso: ME<eacute>todo estE<aacute>tico privado que comprueba si un
fichero con unos resultados dependendientes en un fichero fuente
estE<aacute> desactualizado o no existe.

Devuelve: 1 si hay que crear o actualizar el resultado, 0 si no.

=cut

sub _hay_que_actualizar
{
    my ($paquete, $rutaFuente, $rutaResultado) = @_;
    caller eq __PACKAGE__ or confess;

    return ( ! -f $rutaResultado
             || stat($rutaResultado)->mtime < stat($rutaFuente)->mtime );
}


=head2 _remove_lisp_extensions

Usage: removes the Lisp extensions from a filename.

Returns: original string where the .lsp and .lisp suffixes have been
removed, if any.

We do not remove other extensions, so the output from X.acl2 and
X.lisp will not overwrite each other: we will obtain X.acl2.out and
X.out instead.

=cut
sub _remove_lisp_extensions
{
    my ($filename) = @_;

    foreach my $ext ((".lisp", ".lsp")) {
        $filename = substr $filename, 0, (index $filename, $ext)
            if (index $filename, $ext) != -1;
    }

    return $filename;
}

=head2 _generate_tmp_lisp_file($pathLISP)

Usage: creates a temporary file with the extra lines required to
correctly analyze de output from the original .lisp file, in the
presence of file.acl2 and cert.acl2 files.

Arguments:

=over 2

=item *

C<pathLISP> - path to the original Lisp file to be decorated

=back

Returns: string with the path to the decorated temporary .lisp file.

=cut
sub _generate_tmp_lisp_file
{
    my ($pathLISP)         = @_;
    my $basenameLISP       = _remove_lisp_extensions($pathLISP);
    my ($fh, $tmpFilename) = tempfile()
        or die $lh->maketext("Could not create temporary file: [_1]", $@);

    # 1. Prologue: ensure we always start on the ACL2 REPL
    print $fh
        $lh->maketext("; Prologue added by pprocACL2") . "\n" .
        "(acl2::value :q)\n" .
        "(in-package \"ACL2\")\n" .
        "(acl2::lp)\n\n";

    my $specificBookCert  = "${basenameLISP}.acl2";
    my $directoryBookCert = "cert.acl2";

    # 2a. If a certification script specific to this book exists and
    # is readable:
    use filetest 'access';
    if (-r $specificBookCert) {
        print $fh
            $lh->maketext("; Body dumped from [_1]",
                          File::Spec->rel2abs($specificBookCert)) . "\n" .
            _get_file_contents($specificBookCert) . "\n\n";
    }
    # 2b. Else if a certification script for all books in this
    # directory exists and is readable:
    elsif (-r $directoryBookCert) {
        print $fh
            $lh->maketext("; Body dumped from [_1]",
                          File::Spec->rel2abs($directoryBookCert)) . "\n" .
            _get_file_contents($directoryBookCert) . "\n\n" .
            "; Book certification (added by pprocACL2) \n" .
            "(certify-book \"${basenameLISP}\" ?)\n\n";
    }
    # 2c. Otherwise, we can assume that it is not a book:
    else {
        print $fh
            $lh->maketext("; Body dumped from [_1]",
                          File::Spec->rel2abs($pathLISP)) . "\n" .
            _get_file_contents($pathLISP) . "\n";
    }

    # 3. Epilogue: ensure we always exit both ACL2 and the raw Lisp
    # REPLs by adding :q and (quit). End-of-file does not seem to be
    # enough for some Lisp implementations (old versions of CLisp, for
    # instance).
    print $fh
        $lh->maketext("; Epilogue added by pprocACL2") . "\n" .
        ":q\n" .
        "(quit)\n";

    close($fh);

    return $tmpFilename;
}

=head2 _get_file_contents($path)

Usage: returns the whole contents of a file as a string.

Arguments:

=over 2

=item *

C<path> - path to the file to be read.

=back

=cut
sub _get_file_contents
{
    my ($path) = @_;

    local *FILE;
    open(FILE, $path)
        or die $lh->maketext("No se pudo abrir [_1]: [_2]", $path, $@);
    my $contentsCertScript = join '', <FILE>;
    close(FILE);

    return $contentsCertScript;
}

1;
